#!/usr/bin/env python
# This program is a wrapper around puppet/
# It allows you to configure machine outside of the datacentre
#
# Exit codes:
#     0  all operations succeeded
#     2  invalid module
#    10  system error

import os
import sys
import logging
import argparse
import re
import yaml
import importlib
import subprocess


from locmap.config import LocmapConfig
from locmap.models import Host  # , User
from locmap.logger import configure_logging
from locmap.utils import is_user_admin


def parse_cmdline_args():
    parser = argparse.ArgumentParser(
        description='''
        ** Managing local configuration **,
        ''')
    parser.add_argument('--config_dir', action='store', dest='config_dir',
                        help='Override default config directory')
    parser.add_argument('--environment', action='store', dest='environment',
                        help='Override default puppet environment')
    parser.add_argument('--verbose', action='store_true',
                        help='Increase output verbosity.')
    action = parser.add_mutually_exclusive_group(required=True)
    action.add_argument('--info', action='store_true',
                        help='Print gathered information' +
                             ' about the host and exit',
                        default=False)
    action.add_argument('--list', action='store_true',
                        help='List all available puppet modules')
    action.add_argument('--configure', action='store', dest='configure',
                        metavar='MODULE',
                        help='Run specific puppet module against the system \
                              If - all - run all enabled puppet modules')
    action.add_argument('--enable', action='store', dest='enable',
                        metavar='MODULE',
                        help='Enable a specific module, only enabled modules \
                               can run against the system. (e.g --enable ssh)')
    action.add_argument('--disable', action='store', dest='disable',
                        metavar='MODULE',
                        help="Disable a specific module, disabled modules \
                               can't be configured. (e.g --disable ssh)")
    args = parser.parse_args()
    return args


def main():
    """Application entrypoint"""

    args = parse_cmdline_args()
    configure_logging(args)

    if not is_user_admin():
        logging.error("You should run locmap as root.\n")
        exit(0)

    lconfig = LocmapConfig()
    lconfig.read_config_and_override_with_pargs(args)

    host = Host()
    prepare_locmap_config_layout(lconfig)

    if args.enable is not None:
        retcode = modify_module(lconfig, args.enable, 'enable')
        if retcode == 0:
            print "\n[INFO] module %s %sd.\n" % (args.enable, 'enable')
        else:
            print "\n[ERROR] module %s does not exist.\n" % (args.enable)
        exit(retcode)
    elif args.disable is not None:
        retcode = modify_module(lconfig, args.disable, 'disable')
        if retcode == 0:
            print "\n[INFO] module %s %sd.\n" % (args.disable, 'disable')
        else:
            print "\n[ERROR] module %s does not exist.\n" % (args.disable)
        exit(retcode)
    elif args.list:
        modules = list_modules(lconfig, organisation="linuxsupport",
                               onlyenabled=False)
        print_modules(lconfig, modules)
        exit(0)

    run_input_plugins(lconfig, host)

    if args.info:
        print_host_information(lconfig, host)
        exit(0)
    elif args.configure is not None:
        modules = set()
        if args.configure == "all":
            modules = list_modules(lconfig, organisation="linuxsupport")
        else:
            modules.add(args.configure)

        retcode = configure_module(lconfig, host, modules,
                                   extra_args="", version=4)
        if retcode == 0:
            msg = 'The run succeeded with no changes or failures;'
            msg += ' the system was already in the desired state.'
        elif retcode == 1:
            msg = 'The run failed, or was not attempted'
            msg += ' due to another run already in progress'
        elif retcode == 2:
            msg = 'The run succeeded, and some resources were changed.'
        elif retcode == 4:
            msg = 'The run succeeded, and some resources failed.'
        elif retcode == 6:
            msg = 'The run succeeded, and included both changes and failures.'
        else:
            msg = 'Could not execute puppet.'
        logging.info(msg)
        exit(retcode)


def run_input_plugins(config, hostname):
    from pkg_resources import iter_entry_points
    input_plugin_found = 0
    available_methods = []
    for entry_point in iter_entry_points(group='locmap.plugin', name=None):
        try:
            available_methods.append(entry_point.load())
        except ImportError:
            if config.verbose:
                logging.debug("Failed to load entry_point %s" % entry_point)
            continue
        if config.verbose:
            logging.info("Running input plugin : %s" % entry_point)
    for method in available_methods:
        if "retrieve_info" in method.func_name:
            input_plugin_found += 1
            method(hostname)


def print_user_help(hostname):
    print "\nNo users have been found by the input plugins."
    if hostname.fqdn_contains('cern'):
        print "Please use `useraddcern` to add a CERN user. `man useraddcern`"
    else:
        print "Please use `useradd` to add a local user. `man useradd`"


def print_host_information(config, hostname):
    print "--------------\nConfiguration:\n--------------"
    print "    Configuration directory  : " + config.config_dir
    print "    Configuration file       : " + config.config_file
    print "    Puppet modules path      : " + config.modules_path
    print "    Puppet config path       : " + config.puppet_path
    print "    Puppet hiera path        : " + config.hiera_path
    print "    Puppet environment       : " + config.env
    print "    Log directory            : " + config.log_dir
    print "\n---------\nHostname:\n---------"
    print hostname
    print "\n------\nUsers:\n------"
    users = sorted(hostname.users, key=lambda user: user.login)
    if users:
        for user in users:
            print user
    else:
        print_user_help(hostname)
    sys.exit(0)


def modify_module(config, module, action):
    """
    Manage module status
    """
    is_valid = False
    is_enabled = False
    retcode = 0

    is_enabled = os.path.islink(os.path.join(config.config_dir,
                                             "enabled",
                                             module))
    is_valid = os.path.exists("%s/%s" % (config.modules_path, module))

    target = os.path.join(config.config_dir, "enabled", module)
    source = os.path.join(config.modules_path, module)

    if not is_valid:
        retcode = 3
        return retcode

    if action == "enable" and not is_enabled and is_valid:
        try:
            os.symlink(source, target)
        except Exception as error:
            retcode = 1
            logging.warn("Could not create symlink %s: %s", target, error)
    elif action == "disable" and is_enabled and is_valid:
        try:
            os.remove(target)
        except Exception as error:
            retcode = 1
            logging.warn("Could not delete symlink %s: %s", target, error)
    else:
        retcode = 2

    if (action == "enable" and is_enabled) \
            or (action == "disable" and not is_enabled):
        retcode = 0

    return retcode


def list_modules(config, supporting_module=False,
                 organisation=None, onlyenabled=True):
    """
    Returns the available and enabled puppet modules
    """
    all_modules = set()
    modules = set()

    if not os.path.exists(config.modules_path):
        logging.debug("Puppet module path %s does not exist.",
                      config.modules_path)
        return set()

    all_modules = set([d for d in os.listdir(config.modules_path)
                       if os.path.isdir(os.path.join(config.modules_path, d))])

    for m in all_modules:
        if not supporting_module:
            file = os.path.join(config.modules_path, m, "supporting_module")
            if not os.path.isfile(file):
                modules.add(m)
        else:
            modules.add(m)

    not_organisation = set()
    if organisation is not None:
        for m in modules:
            file = os.path.join(config.modules_path, m, organisation)
            if not os.path.isfile(file):
                not_organisation.add(m)

    not_enabled = set()
    if onlyenabled:
        for m in modules:
            if not os.path.exists(os.path.join(config.config_dir,
                                               'enabled',
                                               m)):
                not_enabled.add(m)

    logging.debug("module list                      : %s" % modules)
    logging.debug("module list without organisation : %s" % not_organisation)
    logging.debug("module list not enabled          : %s" % not_enabled)
    return modules - not_organisation - not_enabled


def print_modules(config, modules):
    print "[Available Modules]"
    template = "{0:15}{1:20}"
    if modules:
        for module in sorted(modules):
            if os.path.exists(config.config_dir + "/enabled/" + module):
                print template.format(module,
                                      "[\033[92m"+" enabled"+"\033[0m]")
            else:
                print template.format(module,
                                      "[\033[91m"+"disabled"+"\033[0m]")
    else:
        logging.warn("No supported module available")


def configure_module(config, host, modules, extra_args="", version=3):
    """
    Run puppet and return execution code
    """
    if config.verbose:
        extra_args = '--test'

    module_list = modules.copy()
    # Handle single module errors case, as multiple module list is curated
    for module in module_list:
        if not os.path.exists("%s/%s" % (config.modules_path, module)):
            modules.remove(module)
            logging.error("Module %s is not installed." % module)
            return 10
        if not os.path.islink(config.config_dir + "/enabled/" + module):
            modules.remove(module)
            logging.error("Module %s is disabled" % module)
            logging.error("Enable it first with `--enable %s`" % module)
            return 11

    site_pp = os.path.join(config.config_dir, "site.pp")
    hiera_config = os.path.join(config.config_dir, "hiera.yaml")
    yaml_dir = os.path.join(config.hiera_path, config.env, "hieradata")
    yaml_file = os.path.join(yaml_dir, "common.yaml")

    # If puppet version 3 FIXME (disable in the future)
    if version == 3:
        cmd = config.puppet_bin + ' apply --disable_warnings deprecations \
           %s \
           --hiera_config=%s \
           --color=false \
           %s \
           --no-stringify_facts'
        to_run = cmd % (site_pp, hiera_config, extra_args)

    elif version == 4:
        cmd = config.puppet_bin + ' apply \
           %s \
           --hiera_config=%s \
           --color=false \
           --modulepath=%s \
           --environment=%s \
           --detailed-exitcodes \
           %s'
        to_run = cmd % (site_pp, hiera_config,
                        config.modules_path, config.env, extra_args)

    logging.debug('puppet command: %s', to_run)

    for module in module_list:
        imported_module = None
        try:
            imported_module = importlib.import_module(
                'locmap.modules.' + module)
            logging.debug("Imported locmap.modules.%s class", module)
        except ImportError:
            logging.debug("No locmap.modules.%s class", module)

        if imported_module:
            try:
                if imported_module.check_requirements(config, host) != 0:
                    modules.remove(module)
                    logging.debug("Requirement failed. Skipping module %s",
                                  module)
            except:
                logging.debug("No check_requirements function for module %s",
                              module)
            try:
                if imported_module.deploy(config, host) == 0:
                    logging.debug("deployed module %s", module)
                else:
                    modules.remove(module)
                    logging.debug("Deployement failed. Skipping module %s",
                                  module)
            except:
                logging.debug("No deploy function for module %s",
                              module)

    with open(yaml_file, 'w') as common:
        mymodules = {'modules_to_run': list(modules)}
        common.write(yaml.dump(
            mymodules,
            explicit_start=True,
            default_flow_style=False))

    host.write_to_yaml(yaml_file)

    return run_and_show_progressbar(to_run, sleep=0,
                                    msg="[INFO    ]" +
                                    " Please wait while your system" +
                                    " is being configured")


def check_puppet_run_exit_code(retcode):
    """
    Provide extra information about the run via exit codes.
    If --detailed-exitcodes 'puppet apply' will use the following exit codes:

    0: The run succeeded with no changes or failures; the system was already in
    the desired state.

    1: The run failed.

    2: The run succeeded, and some resources were changed.

    4: The run succeeded, and some resources failed.

    6: The run succeeded, and included both changes and failures
    """
    pass


def run_and_show_progressbar(cmd, sleep=0, msg='Please wait '):

    # Adding a progress bar if available
    try:
        from progress.spinner import Spinner
        check_progress = True
    except ImportError:
        check_progress = False
        msg = msg + "..."

    # strip extra space
    cmd = re.sub(' +', ' ', cmd)
    msg = msg + ' '

    if sleep > 0:
        cmd = "/usr/bin/sleep %ds && %s" % (sleep, cmd)

    logging.debug("cmd: `%s`" % cmd)

    try:
        p = subprocess.Popen(cmd,
                             stderr=subprocess.PIPE,
                             shell=True,
                             stdout=subprocess.PIPE)

    except:
        logging.info("Could not execute %s\n" % cmd)
        return 10

    if check_progress:
        spinner = Spinner(msg)
        while (p.poll() is None):
            spinner.next()
        print '\n'
    else:
        print msg

    out, err = p.communicate()
    logging.debug("`%s` returns : %s\n" % (cmd, p.returncode))
    logging.debug("`%s` stdout  :\n%s\n" % (cmd, out))
    logging.debug("`%s` stderr  :\n%s\n" % (cmd, err))
    return p.returncode


def prepare_locmap_config_layout(config):
    retcode = 0
    all_dirs = set()
    # logs = []
    all_dirs.add(config.modules_path)
    return retcode

if __name__ == "__main__":
    main()
