from pkg_resources import iter_entry_points
from ConfigParser import SafeConfigParser
from locmap.utils import is_user_admin, merge_two_dicts
from locmap.common import configure_logging
from locmap.config import LocmapConfig

import logging
import ConfigParser
import pkg_resources
import os
import sys
import argparse
import pprint
import yaml
import importlib

def list_enabled_modules(config_dir, modules):
    enabled = set()

    for m in modules:
        path = config_dir + "enabled/" + m
        if os.path.exists(path):
            enabled.add(m)
    return enabled

def modules_inventory(puppet_modules_path, supporting_module=False):
    """
    Return the available and enabled puppet modules
    """
    all_modules = set()
    modules = set()

    if not os.path.exists(puppet_modules_path):
        logging.debug("Puppet module path %s does not exist.", puppet_modules_path)
        return None

    all_modules = set([d for d in os.listdir(puppet_modules_path)
                      if os.path.isdir(os.path.join(puppet_modules_path, d))])

    for m in all_modules:
        if not supporting_module:
            file = puppet_modules_path + m + "/supporting_module"
            if not os.path.isfile(file):
                modules.add(m)
        else:
            modules.add(m)

    return modules

def print_modules(config_dir, modules):
    print "\n[Available Modules]"
    template = "{0:15}{1:20}"
    for module in modules:
        if os.path.exists(config_dir + "/enabled/" + module):
            print template.format(module, "[\033[92m"+"enabled"+"\033[0m]")
        else:
            print template.format(module, "[\033[91m"+"disabled"+"\033[0m]")


def main():
    input = {}
    output = {}
    information = {}
    extra_args = ""

    #FIXME
    #if not is_user_admin():
    #    print '\n[ERROR] User has no admin privileges.\n'
    #   sys.exit(1)

    # locmap arguments
    locmap_parser = argparse.ArgumentParser(
        description='''
        ** Managing puppet modules for Linux local configuration **
        ''',
        epilog='For any problem please contact <linux.support@cern.ch>')
    locmap_parser.add_argument('--configdir', action='store', dest='config_dir',
                               help='Change where config files are found',
                               default='/etc/locmap')
    locmap_parser.add_argument('--info', action='store_true',
                               help='Print gathered information about the \
                               host and its users and exit', default=False),
    locmap_parser.add_argument('--list', action='store_true',
                               help='List all available puppet modules status.')
    locmap_parser.add_argument('--configure', action='store', dest='module_name',
                               help='Run a specific module against the system \
                               or run all modules. (e.g --configure all)')
    locmap_parser.add_argument('--enable', action='store', dest='en_module_name',
                               help='Enable a specific module, only enabled modules \
                               can run against the system. (e.g --enable ssh)')
    locmap_parser.add_argument('--disable', action='store', dest='dis_module_name',
                               help="Disable a specific module, disabled modules \
                               can't be configured. (e.g --disable ssh)")
    locmap_parser.add_argument('--verbose', action='store_true',
                               help='Increase output verbosity.')

    largs = locmap_parser.parse_args()

    if not len(sys.argv) > 1:
        print '\n[ERROR] No option was specified.\n'
        locmap_parser.print_help()
        sys.exit(1)

    if (largs.en_module_name and largs.dis_module_name) \
       or (largs.module_name and largs.dis_module_name) \
       or (largs.info and largs.list) \
       or (largs.info and largs.dis_module_name) \
       or (largs.info and largs.en_module_name) \
       or (largs.list and largs.dis_module_name) \
       or (largs.list and largs.en_module_name) \
       or (largs.module_name and largs.en_module_name):
        print '\n[ERROR] Conflicting options were specified.\n'
        locmap_parser.print_help()
        sys.exit(1)
    if largs.config_dir:
        config_dir = largs.config_dir

    config_dir, config_file, puppet_modules_path, puppet_path, log_dir = \
        read_config(config_dir=config_dir)

    # Create logging handlers

    logs, retcode = prepare_locmap_config_layout(config_dir, config_file,
                                    puppet_modules_path, puppet_path,
                                    log_dir)
    if retcode != 0:
        print "[ERROR] locmap configuration layout not correct."
        print "* Please use --verbose for more information.\n"
        if largs.verbose:
            for log in  logs:
                print "[ERROR] " + log
        sys.exit(1)

    if largs.verbose:
        configure_logging(largs)
        extra_args = "-tv"

    if not os.path.exists(config_dir) \
       or not os.path.exists(puppet_path) \
       or not os.path.isfile(config_file):
        print "\n[ERROR] Specified configuration directories do not exist."
        logging.info("\nconfig_dir:%s\nconfig_file:%s\npuppet_path:%s\n",
                    config_dir, config_file, puppet_path)
        sys.exit(1)
    if largs.list:
        modules =  modules_inventory(puppet_modules_path, supporting_module=False)
        print_modules(config_dir, modules)
        sys.exit(0)

    # Find and execute available input plugin(s)
    input_plugin_found = 0
    available_methods = []

    for entry_point in iter_entry_points(group='locmap.plugin', name=None):
        available_methods.append(entry_point.load())

    for method in available_methods:
        if "retrieve_info" in method.func_name:
            input_plugin_found += 1
            output = method(input)
            information = merge_two_dicts(input, output)
            if(information is None):
                logging.warn("Plugin method did not return a dictionnary\
                            Ignoring output %s from %s",
                            method.func_name,
                            method)
                continue
            input = information

    information = input

    if largs.info:
        logging.info("List of arguments:\n%s", largs)
        logging.info("%s %s %s %s %s",config_dir, config_file,
                    puppet_modules_path, puppet_path, log_dir)
        logging.info("Found %i input plugin(s)", input_plugin_found)
        logging.info("Gathered information: %s", information)
        print "\n------"
        print "--- Configuration directory  : " + config_dir
        print "--- Configuration file       : " + config_file
        print "--- Puppet modules path      : " + puppet_modules_path
        print "--- Puppet config path       : " + puppet_path
        print "--- Log directory            : " + log_dir
        print "------"
        print "--- Number of input plugin found: %d" % input_plugin_found
        print "--- Information gathered by input plugins: "
        pprint.pprint(information)
        print "------\n"
        sys.exit(0)

    if largs.en_module_name or largs.dis_module_name:

        if largs.en_module_name:
            module = largs.en_module_name
            action = "enable"

        if largs.dis_module_name:
            module = largs.dis_module_name
            action = "disable"

        if modify_module(config_dir,
                         puppet_modules_path,
                         module,
                         action) == 0:
            print "\n[INFO] module %s %sd.\n" % (module, action)
            sys.exit(0)
        else:
            print "[ERROR] Could not %s module %s.\n" % (action, module)
            sys.exit(1)

    if largs.module_name:
        modules = set()
        if  largs.module_name == "all":
            modules = modules_inventory(puppet_modules_path)
        else:
            modules.add(largs.module_name)

        modules = list_enabled_modules(config_dir, modules)
        if modules:
            configure(config_dir, puppet_modules_path,
                      puppet_path, modules, information,
                      extra_args)
        else:
            print "[ERROR] No enabled module to configure.\n"
            print "* Please use --list for more information.\n"
            sys.exit(1)

        sys.exit(0)

def configure(config_dir, puppet_modules_path, puppet_path, modules,
              hostinfo, extra_args=""):
    """
    Run puppet and return execution code
    """
    puppet_cmd = 'puppet apply --disable_warnings deprecations %s/site.pp \
            --hiera_config=%s/hiera.yaml \
            --color=false \
            %s \
            --no-stringify_facts'

    yaml_dir = puppet_path + "hieradata/"
    yaml_file = yaml_dir + "common.yaml"

    with open(yaml_file, 'w') as common:
        mymodules = {'modules_to_run': modules}
        common.write(yaml.dump(
            mymodules,
            explicit_start=True,
            default_flow_style=False))

    for module in modules:
        imported_module = None
        try:
            imported_module = importlib.import_module(
                'locmap.modules.' + module)
        except ImportError:
            logging.info("No locmap.modules.%s class", module)

        if imported_module:
            try:
                if imported_module.check_requirements(hostinfo) != 0:
                    logging.info("Requirement failed. Skipping module")
                    continue
            except:
                logging.info("No check_requirements function in module %s",
                            module)
            try:
                imported_module.deploy(hostinfo)
            except:
                logging.info("No deploy function in module %s", module)
    puppet_run = puppet_cmd % (puppet_modules_path, config_dir, extra_args)
    try:
        exe = os.popen(puppet_run).read()
        return 0
    except:
        logging.info("Could not execute puppet for module(s) %s: \n%s\n",
                    modules)
        return 1

def logs(comm, module, logdir):
    """
    Create handler for the puppet modules logs
    """
    today = datetime.date.today()
    formatted_date = today.strftime('Applied in %d, %b %Y')
    #Create the logdir if it's not there.
    if not os.path.exists(logdir):
        os.makedirs(logdir)
    if module == 'all':
        modules_logname = '%s/configure-%s' %(logdir, module)
    else:
        modules_logname = '%s/puppet-%s' %(logdir, module)

    handler = logging.handlers.RotatingFileHandler(
        modules_logname, maxBytes=897882, backupCount=5)
    module_logger = logging.getLogger('Logger for locmap module(s)')
    module_logging.setLevel(logging.DEBUG)
    module_logging.addHandler(handler)
    module_logging.debug(formatted_date + '\n' + comm)


def prepare_locmap_config_layout(config_dir,
                                 config_file,
                                 puppet_modules_path,
                                 puppet_path,
                                 log_dir):

    retcode = 0
    logs = []
    all_dirs = set()
    # Try to create logging directory
    all_dirs.add(puppet_modules_path)
    all_dirs.add(config_dir + "enabled/")
    all_dirs.add(config_dir + "backup/")
    all_dirs.add(puppet_path + "hieradata/module_names/")
    all_dirs.add(puppet_path + "userdata/module_names/")
    all_dirs.add(log_dir)

    for dir in all_dirs:
        if not os.path.exists(dir):
            try:
                os.makedirs(dir, 0755)
            except Exception as error:
                retcode += 1
                logs.append("Could not create %s: %s" % (dir, error))

    for module in modules_inventory(puppet_modules_path):
        hiera_dir = puppet_path + "hieradata/module_names/" + module
        if not os.path.exists(hiera_dir):
            try:
                os.makedirs(hiera_dir)
            except Exception as error:
                retcode += 1
                logs.append("Could not create %s: %s" % (hiera_dir, error))
                user_dir = puppet_path + "userdata/module_names/" + module
                if not os.path.exists(user_dir):
                    try:
                        os.makedirs(user_dir)
                    except Exception as error:
                        logs.append("Could not create %s: %s" % (user_dir, error))
                        retcode += 1

    return logs, retcode


def modify_module(config_dir, puppet_modules_path, module, action):
    """
    Manage module status
    """
    is_valid = False
    is_enabled = False
    retcode = 0

    is_enabled = os.path.islink(config_dir + "enabled/" + module)
    is_valid = os.path.exists("%s%s" % (puppet_modules_path, module))

    target = config_dir + "enabled/" + module
    source = puppet_modules_path + module

    if action == "enable" and not is_enabled:
        try:
            os.symlink(source, target)
        except Exception as error:
            retcode = 1
            logging.warn("Could not create symlink %s: %s", target, error)
    elif action == "disable" and is_enabled:
        try:
            os.remove(target)
        except Exception as error:
            retcode = 1
            logging.warn("Could not delete symlink %s: %s", target, error)
    else:
        retcode = 2

    if (action == "enable" and is_enabled) \
       or (action == "disable" and not is_enabled):
        retcode = 0

    logging.info("Module %s with action %s is valid:%s enabled:%s \
                and return code is %s",
                module,
                action,
                is_valid,
                is_enabled,
                retcode)

    return retcode

def read_config(config_dir="/etc/locmap",
                config_file="locmap.conf",
                puppet_modules_path="/usr/share/puppet/modules",
                puppet_path="/etc/puppet",
                log_dir="/var/log/locmap"):
    """
    Read locmap configuration file values
    """

    # Command to invoke puppet modules
    parser = SafeConfigParser()
    config_dir = os.path.join(config_dir, '')
    config_file = config_dir + config_file
    if os.path.isfile(config_file):
        try:
            config = parser.read(config_file)
            try:
                log_dir = parser.get('general', 'log_dir')
            except ConfigParser.NoOptionError:
                pass
            try:
                puppet_module_path = parser.get('general', 'puppet_module_path')
            except ConfigParser.NoOptionError:
                pass
            try:
                puppet_path = parser.get('general', 'puppet_path')
            except ConfigParser.NoOptionError:
                pass
        except (ConfigParser.NoSectionError,
                ConfigParser.MissingSectionHeaderError):
            pass

    # Format directory with trailing slash
    log_dir = os.path.join(log_dir, '')
    puppet_modules_path = os.path.join(puppet_modules_path, '')
    puppet_path = os.path.join(puppet_path, '')

    return config_dir, config_file, puppet_modules_path, puppet_path, log_dir

